"""
This example can be used to check if the VTK ordering and the ones implemented
in :mod:`pyvisfil.vtk.vtk_ordering` match. It can be useful for debugging
and implementing additional element types.

To facilitate this comparison, and unlike the rest of the package, this example
makes use of the VTK Python bindings.
"""

import numpy as np
import numpy.linalg as la

import matplotlib.pyplot as plt

try:
    import vtk
    from vtkmodules.util.numpy_support import vtk_to_numpy
except ImportError:
    raise ImportError("python bindings for VTK cannot be found")


VTK_LAGRANGE_SIMPLICES = [
        "VTK_LAGRANGE_CURVE",
        "VTK_LAGRANGE_TRIANGLE",
        "VTK_LAGRANGE_TETRAHEDRON",
        ]

VTK_LAGRANGE_QUADS = [
        "VTK_LAGRANGE_QUADRILATERAL",
        "VTK_LAGRANGE_HEXAHEDRON",
        ]


def plot_node_ordering(filename, points, show=False):
    if points.shape[0] == 1:
        points = np.hstack([points, np.zeros_like(points)])

    if points.shape[0] == 2:
        fig = plt.figure(figsize=(8, 8), dpi=300)
        ax = fig.gca()
        ax.plot(points[0], points[1], "o")

        ax.set_xlim([-0.1, 1.1])
        ax.set_xlabel("$x$")
        ax.set_ylim([-0.1, 1.1])
        ax.set_ylabel("$y$")
    elif points.shape[0] == 3:
        from mpl_toolkits.mplot3d import art3d      # noqa: F401

        fig = plt.figure(figsize=(8, 8), dpi=300)
        ax = fig.gca(projection="3d")
        ax.plot(points[0], points[1], points[2], "o-")

        ax.set_xlim([-0.1, 1.1])
        ax.set_xlabel("$x$")
        ax.set_ylim([-0.1, 1.1])
        ax.set_ylabel("$y$")
        ax.set_zlim([-0.1, 1.1])
        ax.set_zlabel("$z$")

        ax.view_init(15, 45)
    else:
        raise ValueError("dimension not supported: %d" % points.shape[0])

    ax.grid()
    for i, p in enumerate(points.T):
        ax.text(*p, str(i), color="k", fontsize=12)

    print("output: %s.png" % filename)
    fig.savefig(filename)
    if show:
        plt.show(block=True)


def create_sample_element(cell_type, order=3, visualize=True):
    from distutils.version import LooseVersion
    if LooseVersion(vtk.VTK_VERSION) > "8.2.0":
        vtk_version = (2, 2)
    else:
        vtk_version = (2, 1)

    cell_type = cell_type.upper()
    vtk_cell_type = getattr(vtk, cell_type, None)
    if vtk_cell_type is None:
        raise ValueError("unknown cell type: '%s'" % cell_type)

    source = vtk.vtkCellTypeSource()
    source.SetCellType(vtk_cell_type)
    source.SetBlocksDimensions(1, 1, 1)
    if "LAGRANGE" in cell_type:
        source.SetCellOrder(order)

    # 0 - single precision; 1 - double precision
    source.SetOutputPrecision(1)
    source.Update()

    grid = source.GetOutput()
    cell = grid.GetCell(0)
    points = vtk_to_numpy(cell.GetPoints().GetData()).copy().T

    dim = cell.GetCellDimension()
    points = points[0:dim]

    basename = f"sample_{cell_type.lower()}"
    if visualize:
        filename = f"{basename}.vtu"

        print("vtk xml version:", vtk_version)
        print("cell type: %s" % cell_type)
        print("output: %s" % filename)

        writer = vtk.vtkXMLUnstructuredGridWriter()
        writer.SetFileName(filename)
        writer.SetCompressorTypeToNone()
        writer.SetDataModeToAscii()
        writer.SetInputData(grid)
        writer.Write()

    # NOTE: vtkCellTypeSource always tesselates a square and the way it does
    # that to get tetrahedra changed in
    #   https://gitlab.kitware.com/vtk/vtk/-/merge_requests/6529
    if LooseVersion(vtk.VTK_VERSION) > "8.2.0" \
            and cell_type == "VTK_LAGRANGE_TETRAHEDRON":
        rot = np.array([
            [1, -1, 0],
            [0, 1, -1],
            [0, 0, 2]
            ])
        points = rot @ points

    if cell_type in VTK_LAGRANGE_SIMPLICES:
        from pyvisfile.vtk.vtk_ordering import (
                vtk_lagrange_simplex_node_tuples,
                vtk_lagrange_simplex_node_tuples_to_permutation)

        node_tuples = vtk_lagrange_simplex_node_tuples(dim, order,
            vtk_version=vtk_version)
        vtk_lagrange_simplex_node_tuples_to_permutation(node_tuples)

        nodes = np.array(node_tuples) / order
        error = la.norm(nodes - points.T)
    elif cell_type in VTK_LAGRANGE_QUADS:
        from pyvisfile.vtk.vtk_ordering import (
                vtk_lagrange_quad_node_tuples,
                vtk_lagrange_quad_node_tuples_to_permutation)

        node_tuples = vtk_lagrange_quad_node_tuples(dim, order,
            vtk_version=vtk_version)
        vtk_lagrange_quad_node_tuples_to_permutation(node_tuples)

        nodes = np.array(node_tuples) / order
        error = la.norm(nodes - points.T)
    else:
        error = None

    # NOTE: skipping the curve check because the ordering is off in the
    # vtkCellTypeSource output
    #   https://gitlab.kitware.com/vtk/vtk/-/merge_requests/6555
    if LooseVersion(vtk.VTK_VERSION) <= "8.2.0" \
            and cell_type == "VTK_LAGRANGE_CURVE":
        error = None

    if error is not None:
        if error < 5.0e-15:
            print(f"\033[92m[PASSED] order {order:2d} error {error:.5e}\033[0m")
        else:
            print(f"\033[91m[FAILED] order {order:2d} error {error:.5e}\033[0m")

    if not visualize:
        return

    filename = f"{basename}_vtk"
    plot_node_ordering(filename, points, show=False)

    if cell_type in (VTK_LAGRANGE_SIMPLICES + VTK_LAGRANGE_QUADS):
        filename = f"{basename}_pyvisfile"
        plot_node_ordering(filename, nodes.T, show=False)


def test_order(max_order=10):
    for cell_type in VTK_LAGRANGE_SIMPLICES:
        print("cell_type:", cell_type)
        for order in range(1, max_order + 1):
            create_sample_element(cell_type, order=order, visualize=False)

    for cell_type in VTK_LAGRANGE_QUADS:
        print("cell_type:", cell_type)
        for order in range(1, max_order + 1):
            create_sample_element(cell_type, order=order, visualize=False)


if __name__ == "__main__":
    test_order()
    create_sample_element("VTK_LAGRANGE_TETRAHEDRON", order=3)



# PyVisfile documentation build configuration file, created by
# sphinx-quickstart on Tue Apr 22 23:46:39 2008.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# The contents of this file are pickled, so don't put values in the namespace
# that aren't pickleable (module imports are okay, they're removed automatically).
#
# All configuration values have a default value; values that are commented out
# serve to show the default value.

# If your extensions are in another directory, add it here.
# sys.path.append("some/directory")

# General configuration
# ---------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named "sphinx.ext.*") or your custom ones.
extensions = [
        "sphinx.ext.autodoc",
        "sphinx.ext.intersphinx",
        "sphinx.ext.mathjax",
        "sphinx_copybutton",
        ]

autodoc_typehints = "description"

# https://github.com/sphinx-doc/sphinx/issues/8127
nitpick_ignore = [
    ("py:class", "..")
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General substitutions.
project = "PyVisfile"
copyright = "2010, Andreas Kloeckner"

# The default replacements for |version| and |release|, also used in various
# other places throughout the built documents.
#
# The short X.Y version.
ver_dic = {}
with open("../pyvisfile/__init__.py") as vpy_file:
    version_py = vpy_file.read()

exec(compile(version_py, "../pyvisfile/__init__.py", "exec"), ver_dic)
version = ".".join(str(x) for x in ver_dic["VERSION"])
# The full version, including alpha/beta/rc tags.
release = version

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ""
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%B %d, %Y"

# List of documents that shouldn't be included in the build.
# unused_docs = []

# If true, "()" will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# Options for HTML output
# -----------------------

html_theme = "furo"

html_theme_options = {
        }

# The style sheet to use for HTML and HTML Help pages. A file of that name
# must exist either in Sphinx' static/ path, or in one of the custom paths
# given in html_static_path.
# html_style = "default.css"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# If not "", a "Last updated on:" timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Content template for the index page.
# html_index = ""

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_use_modindex = True

# If true, the reST sources are included in the HTML build as _sources/<name>.
# html_copy_source = True

# Output file base name for HTML help builder.
htmlhelp_basename = "PyVisfiledoc"

# Options for LaTeX output
# ------------------------

# The paper size ("letter" or "a4").
# latex_paper_size = "letter"

# The font size ("10pt", "11pt" or "12pt").
# latex_font_size = "10pt"

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, document class [howto/manual]).
latex_documents = [("index", "pyvisfile.tex",
    "PyVisfile documentation", "Andreas Kloeckner",
    "manual")]

# Additional stuff for the LaTeX preamble.
# latex_preamble = ""

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True

intersphinx_mapping = {
    "https://docs.python.org/3/": None,
    "https://numpy.org/doc/stable/": None,
    "https://documen.tician.de/modepy": None,
}



import numpy as np
import pytest

from pytools.obj_array import make_obj_array


# {{{ test_unstructured_vertex_grid

@pytest.mark.parametrize("ambient_dim", [2, 3])
@pytest.mark.parametrize("dformat", ["xml", "hdf", "binary"])
def test_unstructured_vertex_grid(ambient_dim, dformat, npoints=64):
    """Test constructing a vertex grid with different ways to define the
    points and connectivity.
    """

    # {{{ set up connectivity

    from pyvisfile.xdmf import NumpyDataArray, DataArray, _data_item_from_numpy
    connectivity = np.arange(npoints, dtype=np.uint32)
    points = np.random.rand(ambient_dim, npoints)

    if dformat == "xml":
        connectivity = NumpyDataArray(connectivity, name="connectivity")
        points = NumpyDataArray(points.T, name="points")
    elif dformat in ["hdf", "binary"]:
        if dformat == "hdf":
            cdata = "geometry.h5:/Grid/Connectivity"
            pdata = "geometry.h5:/Grid/Points"
        else:
            cdata = "connectivity.out"
            pdata = "points.out"

        connectivity = DataArray((
            _data_item_from_numpy(connectivity,
                name="connectivity",
                data=cdata),
            ))
        points = DataArray((
            _data_item_from_numpy(points.T,
                name="points",
                data=pdata),
            ))
    else:
        raise ValueError(f"unknown format: '{dformat}'")

    # }}}

    # {{{ set up grids

    from pyvisfile.xdmf import TopologyType
    from pyvisfile.xdmf import XdmfUnstructuredGrid
    grid = XdmfUnstructuredGrid(
            points, connectivity,
            topology_type=TopologyType.Polyvertex,
            name="polyvertex")

    # }}}

    from pyvisfile.xdmf import XdmfWriter
    writer = XdmfWriter((grid,))

    filename = f"test_unstructured_vertex_{dformat}_{ambient_dim}d.xmf"
    writer.write_pretty(filename)

# }}}


# {{{ test_unstructured_simplex_grid

def _simplex_box_connectivity(*, npoints, nelements, nvertices):
    # NOTE: largely copied from meshmode/mesh/generation.py::generate_box_mesh
    ambient_dim = len(npoints)

    point_indices = np.arange(np.prod(npoints)).reshape(npoints)
    connectivity = np.empty((nelements, nvertices), dtype=np.uint32)

    ielement = 0
    from itertools import product
    if ambient_dim == 1:
        raise NotImplementedError
    elif ambient_dim == 2:
        for i, j in product(range(npoints[0] - 1), repeat=ambient_dim):
            a = point_indices[i + 0, j + 0]
            b = point_indices[i + 1, j + 0]
            c = point_indices[i + 0, j + 1]
            d = point_indices[i + 1, j + 1]

            connectivity[ielement + 0, :] = (a, b, c)
            connectivity[ielement + 1, :] = (d, c, b)
            ielement += 2
    elif ambient_dim == 3:
        for i, j, k in product(range(npoints[0] - 1), repeat=ambient_dim):
            a000 = point_indices[i, j, k]
            a001 = point_indices[i, j, k+1]
            a010 = point_indices[i, j+1, k]
            a011 = point_indices[i, j+1, k+1]

            a100 = point_indices[i+1, j, k]
            a101 = point_indices[i+1, j, k+1]
            a110 = point_indices[i+1, j+1, k]
            a111 = point_indices[i+1, j+1, k+1]

            connectivity[ielement + 0, :] = (a000, a100, a010, a001)
            connectivity[ielement + 1, :] = (a101, a100, a001, a010)
            connectivity[ielement + 2, :] = (a101, a011, a010, a001)

            connectivity[ielement + 3, :] = (a100, a010, a101, a110)
            connectivity[ielement + 4, :] = (a011, a010, a110, a101)
            connectivity[ielement + 5, :] = (a011, a111, a101, a110)
            ielement += 6
    else:
        raise NotImplementedError

    assert ielement == nelements

    from pyvisfile.xdmf import NumpyDataArray
    return NumpyDataArray(connectivity, name="connectivity")


@pytest.mark.parametrize("ambient_dim", [2, 3])
def test_unstructured_simplex_grid(ambient_dim, nelements=16):
    """Test constructing a grid with a more complicated topology."""

    from pyvisfile.xdmf import TopologyType
    if ambient_dim == 1:
        topology_type = TopologyType.Polyline
        simplices_per_quad = 1
    if ambient_dim == 2:
        topology_type = TopologyType.Triangle
        simplices_per_quad = 2
    elif ambient_dim == 3:
        topology_type = TopologyType.Tetrahedron
        simplices_per_quad = 6
    else:
        raise ValueError("unsupported dimension")

    # {{{ points and connectivity

    x = np.linspace(-1.0, 1.0, nelements + 1)

    npoints = len(x)
    points = np.empty((ambient_dim,) + (npoints,) * ambient_dim)
    for idim in range(ambient_dim):
        points[idim] = x.reshape((npoints,) + (1,) * (ambient_dim - 1 - idim))

    from pyvisfile.xdmf import NumpyDataArray
    points = NumpyDataArray(points.reshape(ambient_dim, -1).T, name="points")

    from pyvisfile.xdmf import _XDMF_ELEMENT_NODE_COUNT
    connectivity = _simplex_box_connectivity(
            npoints=(npoints,) * ambient_dim,
            nelements=simplices_per_quad * nelements**ambient_dim,
            nvertices=_XDMF_ELEMENT_NODE_COUNT[topology_type]
            )

    # }}}

    # {{{ attributes

    temperature = np.sin(2.0 * np.pi * points.ary[:, 0]) \
            + np.cos(2.0 * np.pi * points.ary[:, 1])
    temperature = NumpyDataArray(temperature, name="temperature")

    velocity = points.ary + np.array([0, 1, 2][:ambient_dim]).reshape(1, -1)
    velocity = NumpyDataArray(velocity, name="velocity")
    vorticity = NumpyDataArray(make_obj_array(velocity.ary), name="vorticity")

    # }}}

    # {{{ write grids

    from pyvisfile.xdmf import XdmfUnstructuredGrid
    grid = XdmfUnstructuredGrid(
            points, connectivity,
            topology_type=topology_type,
            name="simplex")

    grid.add_attribute(temperature)
    grid.add_attribute(velocity)
    grid.add_attribute(vorticity)

    from pyvisfile.xdmf import XdmfWriter
    writer = XdmfWriter((grid,))

    filename = f"test_unstructured_simplex_{ambient_dim}d.xmf"
    writer.write_pretty(filename)

    # }}}


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        pytest.main([__file__])

# vim: fdm=marker



from pyvisfile.silo import SiloFile, DB_READ
db = SiloFile("qmesh.silo", create=False, mode=DB_READ)
print(db.get_toc().qmesh_names)
print(db.get_toc().qvar_names)

qmesh = db.get_quadmesh("meshxy")
print(qmesh.coords)

qvar = db.get_quadvar("value")
print(qvar.vals)
